"""
Package type detection and meta-data extraction
"""
from __future__ import print_function, unicode_literals

import logging
import tarfile

from os import path

from binstar_client.utils.config import PackageType
from binstar_client.inspect_package.conda import inspect_conda_package
from binstar_client.inspect_package import conda_installer
from binstar_client.inspect_package.pypi import inspect_pypi_package
from binstar_client.inspect_package.r import inspect_r_package
from binstar_client.inspect_package.ipynb import inspect_ipynb_package
from binstar_client.inspect_package.env import inspect_env_package

logger = logging.getLogger('binstar.detect')


def file_handler(filename, fileobj, *args, **kwargs):
    return ({}, {'description': ''},
            {'basename': path.basename(filename), 'attrs': {}})


inspectors = {
    PackageType.CONDA: inspect_conda_package,
    PackageType.STANDARD_PYTHON: inspect_pypi_package,
    PackageType.STANDARD_R: inspect_r_package,
    PackageType.NOTEBOOK: inspect_ipynb_package,
    PackageType.ENV: inspect_env_package,
    PackageType.INSTALLER: conda_installer.inspect_package,
    PackageType.FILE: file_handler,
}


def is_environment(filename):
    logger.debug("Testing if environment file ..")
    if filename.endswith('.yml') or filename.endswith('.yaml'):
        return True
    logger.debug("No environment file")


def is_ipynb(filename):
    logger.debug("Testing if ipynb file ..")
    if filename.endswith('.ipynb'):
        return True
    logger.debug("No ipynb file")


def is_project(filename):
    logger.debug("Testing if project ..")

    def is_python_file():
        return filename.endswith('.py')

    def is_directory():
        return path.isdir(filename)

    if is_directory() or is_python_file():
        return True
    logger.debug("Not a project")


def is_conda(filename):
    logger.debug("Testing if conda package ..")
    if filename.endswith('.conda'):
        return True
    
    if filename.endswith('.tar.bz2'):  # Could be a conda package
        try:
            with tarfile.open(filename, mode="r|bz2") as tf:
                for info in tf:
                    if info.name == "info/index.json":
                        break
                else:
                    raise KeyError
        except KeyError:
            logger.debug("Not conda package no 'info/index.json' file in the tarball")
            return False
        else:
            logger.debug("This is a conda package")
            return True
    logger.debug("Not conda package (file ext is not .tar.bz2 or .conda)")


def is_pypi(filename):
    package_type_label = PackageType.STANDARD_PYTHON.label()
    logger.debug("Testing if {} package ..".format(package_type_label))
    if filename.endswith('.whl'):
        logger.debug("This is a {} wheel package".format(package_type_label))
        return True
    if filename.endswith('.tar.gz') or filename.endswith('.tgz'):  # Could be a setuptools sdist or r source package
        with tarfile.open(filename) as tf:
            if any(name.endswith('/PKG-INFO') for name in tf.getnames()):
                return True
            else:
                logger.debug("This is not a {} package (no '/PKG-INFO' in tarball)".format(package_type_label))
                return False

    logger.debug("This is not a {} package (expected .tgz, .tar.gz or .whl)".format(package_type_label))


def is_r(filename):
    logger.debug("Testing if R package ..")
    if filename.endswith('.tar.gz') or filename.endswith('.tgz'):  # Could be a setuptools sdist or r source package
        with tarfile.open(filename) as tf:

            if (
                    any(name.endswith('/DESCRIPTION') for name in tf.getnames()) and
                    any(name.endswith('/NAMESPACE') for name in tf.getnames())
            ):
                return True
            else:
                logger.debug("This not is an R package (no '*/DESCRIPTION' and '*/NAMESPACE' files).")
    else:
        logger.debug("This not is an R package (expected .tgz, .tar.gz).")


def detect_package_type(filename):
    if isinstance(filename, bytes):
        filename = filename.decode('utf-8', errors='ignore')

    if is_conda(filename):
        return PackageType.CONDA
    elif is_pypi(filename):
        return PackageType.STANDARD_PYTHON
    elif is_r(filename):
        return PackageType.STANDARD_R
    elif is_ipynb(filename):
        return PackageType.NOTEBOOK
    elif is_environment(filename):
        return PackageType.ENV
    elif conda_installer.is_installer(filename):
        return PackageType.INSTALLER
    elif is_project(filename):
        return PackageType.PROJECT


def get_attrs(package_type, filename, *args, **kwargs):
    with open(filename, 'rb') as fileobj:
        inspector = package_type.get_from_mapping(inspectors)
        return inspector(filename, fileobj, *args, **kwargs)



import unittest
from ..env import EnvInspector, inspect_env_package

from binstar_client.utils.test.utils import data_dir


class EnvInspectorTestCase(unittest.TestCase):
    def test_package_name(self):
        with open(data_dir('environment.yml')) as fileobj:
            assert EnvInspector('environment.yml', fileobj).name == 'stats'

    def test_version(self):
        with open(data_dir('environment.yml')) as fileobj:
            self.assertIsInstance(EnvInspector('environment.yml', fileobj).version, str)


class InspectEnvironmentPackageTest(unittest.TestCase):
    def test_inspect_env_package(self):
        with open(data_dir('environment.yml')) as fileobj:
            package_data, release_data, file_data = inspect_env_package(
                'environment.yml', fileobj)

        self.assertEqual({
            'name': 'stats',
            'summary': 'Environment file'
        }, package_data)

        self.assertEqual({
            'basename': 'environment.yml',
            'attrs': {}
        }, file_data)


if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import, print_function, unicode_literals

import os
import re
from datetime import datetime

import nbformat

from ..utils.notebook.data_uri import data_uri_from
from ..utils.notebook.inflection import parameterize


def inspect_ipynb_package(filename, fileobj, *args, **kwargs):
    notebook = nbformat.read(fileobj, nbformat.NO_CONVERT)
    summary = notebook.get('metadata', {}).get('summary', 'Jupyter Notebook')
    description = notebook.get('metadata', {}).get('description', 'Jupyter Notebook')

    package_data = {
        'name': re.sub('\-ipynb$', '', parameterize(os.path.basename(filename))),
        'summary': summary,
        'description': description,
    }

    if 'parser_args' in kwargs and kwargs['parser_args'].thumbnail:
        package_data['thumbnail'] = data_uri_from(kwargs['parser_args'].thumbnail)

    release_data = {
        'version': datetime.now().strftime('%Y.%m.%d.%H%M'),
        'summary': summary,
        'description': description,
    }

    file_data = {
        'basename': os.path.basename(filename),
        'attrs': {}
    }

    return package_data, release_data, file_data



from __future__ import unicode_literals

import json

from binstar_client import errors
from binstar_client.scripts.cli import main
from binstar_client.tests.fixture import CLITestCase
from binstar_client.tests.urlmock import urlpatch
from binstar_client.utils.test.utils import data_dir


with open(data_dir('metadata.json'), 'r') as f:
    json_test_data = json.load(f)


package_test_data = {
    'public_attrs': {
        'license': 'custom',
        'name': 'test-package34',
        'summary': 'Python test package for binstar client'
    }
}
release_test_data = {
    'public_attrs': {
        'description': 'longer description of the package',
        'home_page': 'http://github.com/binstar/binstar_pypi',
        'version': '0.3.1'
    }
}


class TestUpdate(CLITestCase):

    @urlpatch
    def test_update_package_from_json(self, urls):
        urls.register(method='HEAD', path='/', status=200)
        update = urls.register(method='PATCH', path='/package/owner/package_name', content=json_test_data, status=200)
        main(['update', 'owner/package_name', data_dir('metadata.json')], False)

        urls.assertAllCalled()
        req = json.loads(update.req.body)
        self.assertEqual(req, json_test_data)

    @urlpatch
    def test_update_package_from_file(self, urls):
        urls.register(method='HEAD', path='/', status=200)
        update = urls.register(method='PATCH', path='/package/owner/package_name',
                               content=package_test_data, status=200)
        main(['update', 'owner/package_name', data_dir('test_package34-0.3.1.tar.gz')], False)

        urls.assertAllCalled()
        req = json.loads(update.req.body)
        self.assertEqual(req, package_test_data)

    @urlpatch
    def test_update_release_from_json(self, urls):
        urls.register(method='HEAD', path='/', status=200)
        update = urls.register(method='PATCH', path='/release/owner/package_name/1.0.0',
                               content=json_test_data, status=200)
        main(['update', 'owner/package_name/1.0.0', data_dir('metadata.json'), '--release'], False)

        urls.assertAllCalled()
        req = json.loads(update.req.body)
        self.assertEqual(req, json_test_data)

    @urlpatch
    def test_update_release_from_file(self, urls):
        urls.register(method='HEAD', path='/', status=200)
        update = urls.register(method='PATCH', path='/release/owner/package_name/1.0.0',
                               content=release_test_data, status=200)
        main(['update', 'owner/package_name/1.0.0', data_dir('test_package34-0.3.1.tar.gz'), '--release'], False)

        urls.assertAllCalled()
        req = json.loads(update.req.body)
        self.assertEqual(req, release_test_data)

    @urlpatch
    def test_update_local_file_not_found(self, urls):
        urls.register(method='HEAD', path='/', status=200)
        urls.register(method='PATCH', path='/package/owner/package_name', content=package_test_data, status=404)

        with self.assertRaises(SystemExit):
            main(['update', 'owner/package_name', data_dir('not_existing.tar.gz')], False)

    @urlpatch
    def test_update_package_not_found(self, urls):
        urls.register(method='HEAD', path='/', status=200)
        urls.register(method='PATCH', path='/package/owner/package_name', content=package_test_data, status=404)

        with self.assertRaises(errors.NotFound):
            main(['update', 'owner/package_name', data_dir('test_package34-0.3.1.tar.gz')], False)

    @urlpatch
    def test_update_release_missing_version(self, urls):
        urls.register(method='HEAD', path='/', status=200)
        urls.register(method='PATCH', path='/release/owner/package_name/1.0.0', content=package_test_data, status=200)
        with self.assertRaises(errors.UserError):
            main(['update', 'owner/package_name', data_dir('test_package34-0.3.1.tar.gz'), '--release'], False)

    @urlpatch
    def test_update_release_not_found(self, urls):
        urls.register(method='HEAD', path='/', status=200)
        urls.register(method='PATCH', path='/release/owner/package_name/1.0.0', content=package_test_data, status=404)

        with self.assertRaises(errors.NotFound):
            main(['update', 'owner/package_name/1.0.0', data_dir('test_package34-0.3.1.tar.gz'), '--release'], False)



from __future__ import print_function, absolute_import, unicode_literals

import collections
import enum
import itertools
import logging
import os
import shutil
import stat
import warnings
from string import Template

import yaml

from binstar_client.errors import BinstarError

try:
    from urllib import quote_plus
except ImportError:
    from urllib.parse import quote_plus

from binstar_client.utils.conda import CONDA_PREFIX, CONDA_ROOT
from binstar_client.utils.appdirs import AppDirs, EnvAppDirs

from .yaml import yaml_load, yaml_dump


logger = logging.getLogger('binstar')


def expandvars(path):
    environ = dict(CONDA_ROOT=CONDA_ROOT, CONDA_PREFIX=CONDA_PREFIX)
    environ.update(os.environ)
    return Template(path).safe_substitute(**environ)


def expand(path):
    return os.path.abspath(os.path.expanduser(expandvars(path)))


if 'BINSTAR_CONFIG_DIR' in os.environ:
    dirs = EnvAppDirs('binstar', 'ContinuumIO', os.environ['BINSTAR_CONFIG_DIR'])
    USER_CONFIG = os.path.join(dirs.user_data_dir, 'config.yaml')
else:
    dirs = AppDirs('binstar', 'ContinuumIO')
    USER_CONFIG = os.path.join(os.path.expanduser('~'), '.continuum', 'anaconda-client', 'config.yaml')


class PackageType(enum.Enum):
    CONDA = 'conda'
    ENV = 'env'
    FILE = 'file'
    NOTEBOOK = 'ipynb'
    STANDARD_PYTHON = 'pypi'
    STANDARD_R = 'r'
    PROJECT = 'project'
    INSTALLER = 'installer'

    def label(self):
        return self.get_from_mapping(PACKAGE_TYPE_LABELS, default=self.value)

    def get_from_mapping(self, mapping, default=None):
        return mapping.get(self, default)

    @classmethod
    def _missing_(cls, value):
        try:
            return cls(PACKAGE_TYPE_ALIASES[value])
        except KeyError:
            return super()._missing_(value)


PACKAGE_TYPE_LABELS = {
    PackageType.ENV: 'Environment',
    PackageType.NOTEBOOK: 'Notebook',
    PackageType.CONDA: 'Conda',
    PackageType.STANDARD_PYTHON: 'Standard Python',
    PackageType.STANDARD_R: 'Standard R',
}

PACKAGE_TYPE_ALIASES = {
    'PyPI': 'pypi',
    'standard_python': 'pypi',

    'cran': 'r',
    'standard_r': 'r',
}


USER_LOGDIR = dirs.user_log_dir
SITE_CONFIG = os.path.join(os.environ.get('CONDA_ROOT', CONDA_ROOT), 'etc', 'anaconda-client', 'config.yaml')
SYSTEM_CONFIG = SITE_CONFIG

DEFAULT_URL = 'https://api.anaconda.org'
DEFAULT_CONFIG = {
    'sites': {
        'anaconda': {'url': DEFAULT_URL},
        'binstar': {'url': DEFAULT_URL},
    },
    'auto_register': True,
    'default_site': None,
    'url': DEFAULT_URL,
    'ssl_verify': True
}

CONFIGURATION_KEYS = [
    'auto_register',
    'default_site',
    'upload_user',
    'sites',
    'url',
    'verify_ssl',
    'ssl_verify',
]

SEARCH_PATH = (
    dirs.site_data_dir,
    '/etc/anaconda-client/',
    '$CONDA_ROOT/etc/anaconda-client/',
    dirs.user_data_dir,
    '~/.continuum/anaconda-client/',
    '$CONDA_PREFIX/etc/anaconda-client/',
)


def recursive_update(config, update_dict):
    for update_key, updated_value in update_dict.items():
        if isinstance(updated_value, collections.abc.Mapping):
            updated_value_dict = recursive_update(config.get(update_key, {}), updated_value)
            config[update_key] = updated_value_dict
        else:
            config[update_key] = update_dict[update_key]

    return config


def get_server_api(token=None, site=None, cls=None, config=None, **kwargs):
    """
    Get the anaconda server api class
    """
    if not cls:
        from binstar_client import Binstar
        cls = Binstar

    config = config if config is not None else get_config(site=site)

    url = config.get('url', DEFAULT_URL)

    logger.info("Using Anaconda API: %s", url)

    if token:
        logger.debug("Using token from command line args")
    elif 'BINSTAR_API_TOKEN' in os.environ:
        logger.debug("Using token from environment variable BINSTAR_API_TOKEN")
        token = os.environ['BINSTAR_API_TOKEN']
    elif 'ANACONDA_API_TOKEN' in os.environ:
        logger.debug("Using token from environment variable ANACONDA_API_TOKEN")
        token = os.environ['ANACONDA_API_TOKEN']
    else:
        token = load_token(url)

    verify = config.get('ssl_verify', config.get('verify_ssl', True))

    return cls(token, domain=url, verify=verify, **kwargs)


def get_binstar(args=None, cls=None):
    """
    DEPRECATED METHOD,

    use `get_server_api`
    """

    warnings.warn(
        'method get_binstar is deprecated, please use `get_server_api`',
        DeprecationWarning
    )

    token = getattr(args, 'token', None)
    log_level = getattr(args, 'log_level', logging.INFO)
    site = getattr(args, 'site', None)

    aserver_api = get_server_api(token=token, site=site, log_level=log_level, cls=cls)
    return aserver_api


TOKEN_DIRS = [
    dirs.user_data_dir,
    os.path.join(os.path.dirname(USER_CONFIG), 'tokens'),
]
TOKEN_DIR = TOKEN_DIRS[-1]


def store_token(token, args):
    config = get_config(site=args and args.site)

    for token_dir in TOKEN_DIRS:
        url = config.get('url', DEFAULT_URL)

        if not os.path.isdir(token_dir):
            os.makedirs(token_dir)
        tokenfile = os.path.join(token_dir, '%s.token' % quote_plus(url))

        if os.path.isfile(tokenfile):
            os.unlink(tokenfile)
        with open(tokenfile, 'w') as fd:
            fd.write(token)
        os.chmod(tokenfile, stat.S_IWRITE | stat.S_IREAD)


def load_token(url):
    for token_dir in TOKEN_DIRS:
        tokenfile = os.path.join(token_dir, '%s.token' % quote_plus(url))

        if os.path.isfile(tokenfile):
            logger.debug("Found login token: {}".format(tokenfile))
            with open(tokenfile) as fd:
                token = fd.read().strip()

            if token:
                return token
            else:
                logger.debug("Token file is empty: {}".format(tokenfile))
                logger.debug("Removing file: {}".format(tokenfile))
                os.unlink(tokenfile)


def remove_token(args):
    config = get_config(site=args and args.site)
    url = config.get('url', DEFAULT_URL)

    for token_dir in TOKEN_DIRS:
        tokenfile = os.path.join(token_dir, '%s.token' % quote_plus(url))
        if os.path.isfile(tokenfile):
            os.unlink(tokenfile)


def load_config(config_file):
    data = {}
    warn_msg = None

    try:
        with open(config_file) as fd:
            data = yaml_load(fd) or data
    except yaml.YAMLError:
        backup_file = config_file + '.bak'
        shutil.copyfile(config_file, backup_file)
        warn_msg = "Config file `{}` has invalid structure and couldn't be read. \n"\
                   "File content was backed up to `{}`".format(config_file, backup_file)
    except PermissionError:
        warn_msg = 'Not enough rights to access config file `{}`! Please review file permissions.'.format(config_file)
    except OSError as error:
        logger.exception(error)

    if warn_msg is not None:
        warnings.warn(warn_msg)

    return data


def load_file_configs(search_path):
    def _file_yaml_loader(fullpath):
        assert fullpath.endswith(".yml") or fullpath.endswith(".yaml") or fullpath.endswith("anacondarc"), fullpath
        yield fullpath, load_config(fullpath)

    def _dir_yaml_loader(fullpath):
        for filename in os.listdir(fullpath):
            if filename.endswith(".yml") or filename.endswith(".yaml"):
                filepath = os.path.join(fullpath, filename)
                yield filepath, load_config(filepath)

    # map a stat result to a file loader or a directory loader
    _loader = {
        stat.S_IFREG: _file_yaml_loader,
        stat.S_IFDIR: _dir_yaml_loader,
    }

    def _get_st_mode(path):
        # stat the path for file type, or None if path doesn't exist
        try:
            return stat.S_IFMT(os.stat(path).st_mode)
        except OSError:
            return None

    expanded_paths = [
        expand(path)
        for path in search_path
    ]
    stat_paths = (
        _get_st_mode(path)
        for path in expanded_paths
    )
    load_paths = (
        _loader[st_mode](path)
        for path, st_mode in zip(expanded_paths, stat_paths)
        if st_mode is not None
    )
    raw_data = collections.OrderedDict(
        kv
        for kv in itertools.chain.from_iterable(load_paths)
    )

    return raw_data


def get_config(site=None):
    config = DEFAULT_CONFIG.copy()

    file_configs = load_file_configs(SEARCH_PATH)
    for fn in file_configs:
        recursive_update(config, file_configs[fn])

    site = site or config.get('default_site')
    sites = config.get('sites', {})

    if site:
        site = str(site)

        if site not in sites:
            logger.warning('Site alias "%s" does not exist in the config file', site)
        else:
            # This takes whatever keys are set for the site into the top level of the config dict
            recursive_update(config, sites.get(site, {}))

    return config


def save_config(data, config_file):
    try:
        os.makedirs(os.path.dirname(config_file), exist_ok=True)

        temp_file = config_file + '~'
        with open(temp_file, 'w') as stream:
            yaml_dump(data, stream=stream)

        os.replace(temp_file, config_file)

    except (OSError, yaml.YAMLError):
        raise BinstarError("Config file `{}` couldn't be saved! Changes may be lost.".format(config_file))


def set_config(data, user=True):
    warnings.warn('Use save_config instead of set_config', DeprecationWarning)
    save_config(data, USER_CONFIG if user else SYSTEM_CONFIG)



'''
Created on Feb 18, 2014

@author: sean
'''

from __future__ import unicode_literals

import json
import mock
import os
import unittest
import requests.utils

from binstar_client.scripts.cli import main
from binstar_client.tests.fixture import CLITestCase
from binstar_client.tests.urlmock import urlpatch
from binstar_client.utils.test.utils import data_dir


class Test(CLITestCase):
    @urlpatch
    def test_whoami_anon(self, urls):
        user = urls.register(method='GET', path='/user', status=401)
        main(['--show-traceback', 'whoami'], False)
        self.assertIn('Anonymous User', self.stream.getvalue())

        user.assertCalled() 

    @urlpatch
    def test_whoami(self, urls):
        content = json.dumps({'login': 'eggs', 'created_at':'1/2/2000'})
        user = urls.register(method='GET', path='/user', content=content)

        main(['--show-traceback', 'whoami'], False)
        self.assertIn('eggs', self.stream.getvalue()) 

        user.assertCalled()

    @urlpatch
    def test_netrc_ignored(self, urls):
        # Disable token authentication
        self.load_token.return_value = None
        os.environ.pop('BINSTAR_API_TOKEN', None)
        os.environ.pop('ANACONDA_API_TOKEN', None)

        # requests.get_netrc_auth uses expanduser to find the netrc file, point to our test file
        expanduser = mock.Mock(return_value = data_dir('netrc'))
        with mock.patch('os.path.expanduser', expanduser):
            auth = requests.utils.get_netrc_auth('http://localhost', raise_errors=True)
        self.assertEqual(auth, ('anonymous', 'pass'))

        user = urls.register(path='/user', status=401)

        main(['--show-traceback', 'whoami'], False)
        self.assertNotIn('Authorization', user.req.headers)


if __name__ == '__main__':
    unittest.main()



from __future__ import print_function, unicode_literals

import unittest

from binstar_client.inspect_package import r
from binstar_client.utils.test.utils import data_dir


expected_package_data = {
    'license': 'GPL-2 | GPL-3',
    'name': 'rfordummies',
    'summary': 'Code Examples to Accompany the Book "R for Dummies"'
}


expected_version_data = {
    'description': 'Contains all the code examples in the book "R for Dummies" (1st\n    edition). '
                   'You can view the table of contents as well as the sample code for each\n    chapter.',
    'version': '0.1.2',
}

expected_file_data = {
    'attrs': {
        'NeedsCompilation': 'no',
        'suggests': [
            'fortunes',
            'stringr',
            'sos',
            'XLConnect',
            'reshape2',
            'ggplot2',
            'foreign',
            'lattice',
        ],
        'depends': [],
        'type': 'source',
    },
    'basename': 'rfordummies_0.1.2.tar.gz',
}


class Test(unittest.TestCase):
    maxDiff = None
    def test_r(self):
        filename = data_dir('rfordummies_0.1.2.tar.gz')
        with open(filename, 'rb') as fd:
            package_data, version_data, file_data = r.inspect_r_package(filename, fd)

        self.assertEqual(expected_package_data, package_data)
        self.assertEqual(expected_version_data, version_data)
        self.assertEqual(expected_file_data, file_data)


if __name__ == "__main__":
    unittest.main()
